#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <algorithm>

#include "cpu_benchmark.h"
#include "log.h"
using std::cout;

void CPUBenchmark::diffusion3d(
  FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
  FLOAT dx, FLOAT dy, FLOAT dz, FLOAT dt, FLOAT kappa) {
  FLOAT ce, cw, cn, cs, ct, cb, cc;
  ce = cw = kappa*dt/(dx*dx);
  cn = cs = kappa*dt/(dy*dy);
  ct = cb = kappa*dt/(dz*dz);
  cc = 1.0 - (ce + cw + cn + cs + ct + cb);

  int j, jx, jy, jz;
  for (jz = 0; jz < nz; jz++) {
    for (jy = 0; jy < ny; jy++) {
      for (jx = 0; jx < nx; jx++) {
        FLOAT e, w, n, s, t, b, c;
        j = jz*nx*ny + jy*nx + jx;
        c = f[j];
        w = (jx == 0)    ? c : f[j-1];
        e = (jx == nx-1) ? c : f[j+1];
        n = (jy == 0)    ? c : f[j-nx];
        s = (jy == ny-1) ? c : f[j+nx];
        b = (jz == 0)    ? c : f[j-nx*ny];
        t = (jz == nz-1) ? c : f[j+nx*ny];
        fn[j] = cc*c
          + cw*w + ce*e
          + cs*s + cn*n
          + cb*b + ct*t;
      }
    }
  }
}

void CPUBenchmark::initialize(const int nx, const int ny, const int nz) {
  Benchmark::initialize(nx, ny, nz);
  Problem *p = &problem_;
  size_t array_size;

  delete[] fore_buffer_;
  delete[] back_buffer_;

  array_size   = p->n.x * p->n.y * p->n.z;
  fore_buffer_ = new FLOAT[array_size];
  back_buffer_ = new FLOAT[array_size];
  memcpy(fore_buffer_, p->f(), array_size*sizeof(FLOAT));
}

void CPUBenchmark::run() {
  int    nx    = problem_.n.x;
  int    ny    = problem_.n.y;
  int    nz    = problem_.n.z;
  FLOAT  dx    = problem_.d.x;
  FLOAT  dy    = problem_.d.y;
  FLOAT  dz    = problem_.d.z;
  FLOAT  dt    = problem_.dt;
  FLOAT  k     = problem_.kappa;
  FLOAT *f     = fore_buffer_;
  FLOAT *fn    = back_buffer_;
  FLOAT  time  = 0.0;
  int    count = 0;
  startTimer();
  do {
    if (count && (count % 100 == 0)) {
      LOG_INFO("time(%4d)=%7.5f", count, time + dt);
    }

    diffusion3d(f, fn, nx, ny, nz, dx, dy, dz, dt, k);
    std::swap(f, fn);
    time += dt;
    count++;
  } while (time + 0.5*dt < 0.1);
  stopTimer();
  result_time_     = time;
  result_accuracy_ = problem_.accuracy(f, time);
  result_flops_    = (nx*ny*nz)*13.0*count/elapsed_time_*1.0e-06;
}
