#include "gpu_benchmark.h"
#include <cuda.h>
#include <cutil.h>
#include "log.h"
#include "cuda_common.h"
#define DIM_X 64
#define DIM_Y 4

__global__
void gpu_diffusion3d(FLOAT *f, FLOAT *fn,
                     int nx, int ny, int nz, FLOAT ce, FLOAT cw,
                     FLOAT cn, FLOAT cs, FLOAT ct, FLOAT cb, FLOAT cc) {
    FLOAT e, w, n, s, t, b, c;
    int jx, jy, jz, j, ns; 
    jx = blockDim.x*blockIdx.x + threadIdx.x;
    jy = blockDim.y*blockIdx.y + threadIdx.y;
    ns = nx*ny;

    j = 0*ns + jy*nx + jx;
    for (jz = 0; jz < nz; jz++) {
        int iw, ie, is, in, ib, it;
        c = f[j];
        iw = (jx == 0     ) ? j : j-1;
        ie = (jx == nx - 1) ? j : j+1;
        is = (jy == 0     ) ? j : j-nx;
        in = (jy == ny - 1) ? j : j+nx;
        ib = (jz == 0     ) ? j : j-ns;
        it = (jz == nz - 1) ? j : j+ns;

        w = f[iw];
        e = f[ie];
        s = f[is];
        n = f[in];
        b = f[ib];
        t = f[it];
        fn[j] = cc*c + cw*w + ce*e + cn*n + cs*s + cb*b + ct*t;
        j += ns;
    }
}

__global__
void gpu_diffusion3d1(
        FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
        FLOAT ce, FLOAT cw, FLOAT cn, FLOAT cs, FLOAT ct, FLOAT cb, FLOAT cc) {
    FLOAT e, w, n, s, t, b, c;
    int jx, jy, jz, j, ns;
    jx = blockDim.x*blockIdx.x + threadIdx.x;
    jy = blockDim.y*blockIdx.y + threadIdx.y;
    ns = nx*ny;

    j = 0*ns + jy*nx + jx;
    for (jz = 0; jz < nz; jz++) {
        c = f[j];
        w = (jx == 0     ) ? c : f[j-1];
        e = (jx == nx - 1) ? c : f[j+1];
        s = (jy == 0     ) ? c : f[j-nx];
        n = (jy == ny - 1) ? c : f[j+nx];
        b = (jz == 0     ) ? c : f[j-ns];
        t = (jz == nz - 1) ? c : f[j+ns];
        fn[j] = cc*c + cw*w + ce*e + cn*n + cs*s + cb*b + ct*t;
        j += ns;
    }
}

__global__
void  gpu_diffusion3d2(
        FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
        FLOAT ce, FLOAT cw, FLOAT cn, FLOAT cs, FLOAT ct, FLOAT cb, FLOAT cc) {
    FLOAT fz_p, fz_c, fz_n;
    int j, jx, jy, jz;
    __shared__ FLOAT fs[DIM_Y+2][DIM_X+2];

    int ix = threadIdx.x, iy = threadIdx.y;
    jx = blockDim.x*blockIdx.x + ix;
    jy = blockDim.y*blockIdx.y + iy;

    j = 0*nx*ny + jy*nx + jx;
    fz_p = fz_c = f[j];
    for (jz = 0; jz < nz; jz++) {
        syncthreads();
        fs[iy+1][ix+1] = fz_c;
        if (threadIdx.x == 0) {
            if (jx == 0)      fs[iy+1][0      ] = fz_c;
            else              fs[iy+1][0      ] = f[j - 1];
        }
        if (ix == DIM_X - 1) {
            if (jx == nx - 1) fs[iy+1][DIM_X+1] = fz_c;
            else              fs[iy+1][DIM_X+1] = f[j + 1];
        }
        if (iy == 0) {
            if (jy == 0)      fs[0      ][ix+1] = fz_c;
            else              fs[0      ][ix+1] = f[j - nx];
        }
        if (iy == DIM_Y - 1) {
            if (jy == ny - 1) fs[DIM_Y+1][ix+1] = fz_c;
            else              fs[DIM_Y+1][ix+1] = f[j + nx];
        }
        if (jz == nz - 1)   fz_n = fz_c;
        else                fz_n = f[j+nx*ny];

        syncthreads();

        fn[j] = ct*fz_n + cc*fz_c + cb*fz_p
                + cw*fs[iy+1][ix  ] + ce*fs[iy+1][ix+2]
                + cs*fs[iy  ][ix+1] + cn*fs[iy+2][ix+1];
        fz_p = fz_c;
        fz_c = fz_n;
        j += nx*ny;
    }
}

GPUBenchmark::~GPUBenchmark() {
    CUDA_SAFE_CALL(cudaFreeHost(host_buffer_));
}

void GPUBenchmark::initialize(const int nx, const int ny, const int nz) {
    Benchmark::initialize(nx, ny, nz);
    Problem *p;
    size_t array_size;
    p          = &problem_;
    array_size = p->n.x*p->n.y*p->n.z*sizeof(FLOAT);

    set_device();
    CUDA_SAFE_CALL(cudaFreeHost(host_buffer_));
    CUDA_SAFE_CALL(cudaMallocHost((void **)&host_buffer_, array_size));
    CUDA_SAFE_CALL(cudaMemcpy(host_buffer_, p->f(), array_size,
                              cudaMemcpyHostToHost));
}

void GPUBenchmark::run() {
    int    nx    = problem_.n.x;
    int    ny    = problem_.n.y;
    int    nz    = problem_.n.z;
    FLOAT  dx    = problem_.d.x;
    FLOAT  dy    = problem_.d.y;
    FLOAT  dz    = problem_.d.z;
    FLOAT  dt    = problem_.dt;
    FLOAT  kappa = problem_.kappa;
    FLOAT *f, *fn;
    FLOAT  time  = 0.0;
    int    count = 0;
    size_t array_size = nx*ny*nz*sizeof(FLOAT);


  
    CUDA_SAFE_CALL(cudaMalloc((void**)&f, array_size));
    CUDA_SAFE_CALL(cudaMalloc((void**)&fn, array_size));
    CUDA_SAFE_CALL(cudaMemcpy(f,host_buffer_,array_size,cudaMemcpyHostToDevice));

    startTimer();
    dim3 grid(nx/DIM_X, ny/DIM_Y, 1);
    dim3 threads(DIM_X, DIM_Y, 1);
    FLOAT ce = kappa*dt/(dx*dx), cw = kappa*dt/(dx*dx),
            cn = kappa*dt/(dy*dy), cs = kappa*dt/(dy*dy),
            ct = kappa*dt/(dz*dz), cb = kappa*dt/(dz*dz),
            cc = 1.0 - (ce + cw + cn + cs + ct + cb);

    // Cache configuration for Fermi-based GPUs
    //cudaFuncSetCacheConfig(gpu_diffusion3d, cudaFuncCachePreferL1);
    //cudaFuncSetCacheConfig(gpu_diffusion3d1, cudaFuncCachePreferL1);
    //cudaFuncSetCacheConfig(gpu_diffusion3d2, cudaFuncCachePreferShared);
    do {
        if (count && (count % 100 == 0)) {
            LOG_INFO("time(%4d)=%7.5f", count, time + dt);
        }

        gpu_diffusion3d<<<grid,threads >>>(f,fn,nx,ny,nz,ce,cw,cn,cs,ct,cb,cc);

        std::swap(f, fn);
        time += dt;
        count++;
    } while (time + 0.5*dt < 0.1);

    CUDA_SAFE_CALL(cudaThreadSynchronize());
    stopTimer();
    CUDA_SAFE_CALL(cudaMemcpy(host_buffer_,f,array_size,cudaMemcpyDeviceToHost));

    result_time_     = time;
    result_accuracy_ = problem_.accuracy(host_buffer_, time);
    result_flops_    = (nx*ny*nz)*13.0*count/elapsed_time_*1.0e-06;

    CUDA_SAFE_CALL(cudaFree(f));
    CUDA_SAFE_CALL(cudaFree(fn));
}
