#include <mpi.h>
#include <iostream>
#include "mpi_benchmark.h"
#include "log.h"

using namespace std;
void MPI_Benchmark::initialize(const int nx, const int ny, const int nz) {
  int name_len;
  char name[MPI_MAX_PROCESSOR_NAME];
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_num_process_);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank_);
  MPI_Get_processor_name(name, &name_len);
  mpi_processor_name_ = name;
  problem_n_ = Size3d(nx, ny, nz);
  problem_.initialize(nx, ny, nz);
  if (mpi_rank_ == 0) {
    int count = nx * ny * nz;
    int bound = nx * ny;
    delete[] problem_f_;
    problem_f_ = new FLOAT[count + bound*2];
    memcpy(problem_f_             , problem_.f(), bound*sizeof(FLOAT));
    memcpy(problem_f_ +bound      , problem_.f(), count*sizeof(FLOAT));
    memcpy(problem_f_ +bound+count, problem_.f()+count-bound,
           bound*sizeof(FLOAT));
    LOG_INFO("sent checksum : %f", getSum(problem_.f(), nx*ny*nz));
  }
}

void MPI_Benchmark::run() {
  MPI_Status status;
  int    nprocs    = mpi_num_process_;
  int    rank      = mpi_rank_;
  int    nx        = problem_n_.x;
  int    ny        = problem_n_.y;
  int    nz        = problem_n_.z;
  int    ns        = nx * ny;
  FLOAT  dx        = problem_.d.x;
  FLOAT  dy        = problem_.d.y;
  FLOAT  dz        = problem_.d.z;
  FLOAT  dt        = problem_.dt;
  FLOAT  k         = problem_.kappa;
  int    dim_z     = nz / nprocs;
  int    rem_z     = nz % nprocs;
  int    count     = ns * dim_z;
  int    rem_count = ns * rem_z;
  int    z_bound   = ns;
  int    nz_self   = rank == 0 ? dim_z + rem_z : dim_z;
  int    size      = ns * (nz_self + 2);
  FLOAT *f         = new FLOAT[size];
  FLOAT *fn        = new FLOAT[size];
  FLOAT  time      = 0.0;
  int    iter_count = 0;

  if (rank == 0) {
    startTimer();
    // copy data for rank0 self.
    memcpy(f, problem_f_, (count + rem_count + z_bound*2)*sizeof(FLOAT));
    for (int i = 1; i < nprocs; i++) {
      MPI_Send(problem_f_ + rem_count + i*count, count + z_bound*2,
               MPI_FLOAT, i, 0, MPI_COMM_WORLD);
    }
  }
  else {
    MPI_Recv(f, count + z_bound*2, MPI_FLOAT,
             0, 0, MPI_COMM_WORLD, &status);
  }

  // run kernel
  do {
    exchangeBoundaries(f, size, z_bound);

    if (iter_count && (iter_count % 100 == 0)) {
      LOG_INFO("time(%4d)=%7.5f", iter_count, time + dt);
    }
    diffusion3d(f, fn, nx, ny, nz_self, dx, dy, dz, dt, k);
    if (rank == 0) {
      memcpy(fn, fn + z_bound, z_bound*sizeof(FLOAT));
    }
    else if (rank == nprocs - 1) {
      memcpy(fn + size - z_bound, fn + size - z_bound*2, z_bound*sizeof(FLOAT));
    }
    std::swap(f, fn);
    time += dt;
    iter_count++;
  } while (time + 0.5*dt < 0.1);

  // send data back to rank0
  if (rank == 0) {
    LOG_INFO("calculation has been done.\n");
    FLOAT *answer = new FLOAT[nx*ny*nz];
    memcpy(answer, f + z_bound, (count+rem_count)*sizeof(FLOAT));
    for (int i = 1; i < nprocs; i++) {
      MPI_Recv(answer + rem_count + i*count, count, MPI_FLOAT,
               i, 0, MPI_COMM_WORLD, &status);
    }
    LOG_INFO("recv checksum: %f", getSum(answer, nx*ny*nz));
    stopTimer();
    result_time_     = time;
    result_accuracy_ = problem_.accuracy(answer, time);
    result_flops_    = (nx*ny*nz)*13.0*iter_count/elapsed_time_*1.0e-06;
    delete[] answer;
  }
  else {
    MPI_Send(f + z_bound, count, MPI_FLOAT,
             0, 0, MPI_COMM_WORLD);
  }
  delete[] f;
  delete[] fn;
}

#define DIM_X 64
#define DIM_Y 4
__global__
void mpi_gpu_diffusion3d(
  FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
  FLOAT ce, FLOAT cw, FLOAT cn, FLOAT cs, FLOAT ct, FLOAT cb, FLOAT cc) {
  FLOAT e, w, n, s, t, b, c;
  int jx, jy, jz, j, ns;
  jx = blockDim.x*blockIdx.x + threadIdx.x;
  jy = blockDim.y*blockIdx.y + threadIdx.y;
  ns = nx*ny;

  j = 1*ns + jy*nx + jx;
  for (jz = 0; jz < nz; jz++) {
    int iw, ie, is, in, ib, it;
    c = f[j];
    iw = (jx == 0     ) ? j : j-1;
    ie = (jx == nx - 1) ? j : j+1;
    is = (jy == 0     ) ? j : j-nx;
    in = (jy == ny - 1) ? j : j+nx;
    ib = j-ns;
    it = j+ns;

    w = f[iw];
    e = f[ie];
    s = f[is];
    n = f[in];
    b = f[ib];
    t = f[it];
    fn[j] = cc*c + cw*w + ce*e + cn*n + cs*s + cb*b + ct*t;
    j += ns;
  }
}

void MPI_GPUBenchmark::run() {
  MPI_Status status;
  int    nprocs    = mpi_num_process_;
  int    rank      = mpi_rank_;
  int    nx        = problem_n_.x;
  int    ny        = problem_n_.y;
  int    nz        = problem_n_.z;
  int    ns        = nx * ny;
  FLOAT  dx        = problem_.d.x;
  FLOAT  dy        = problem_.d.y;
  FLOAT  dz        = problem_.d.z;
  FLOAT  dt        = problem_.dt;
  FLOAT  k         = problem_.kappa;
  int    dim_z     = nz / nprocs;
  int    rem_z     = nz % nprocs;
  int    count     = ns * dim_z;
  int    rem_count = ns * rem_z;
  int    z_bound   = ns;
  int    nz_self   = rank == 0 ? dim_z + rem_z : dim_z;
  int    size      = ns * (nz_self + 2);
  FLOAT *host_buffer;
  FLOAT *f;
  FLOAT *fn;
  CUDA_SAFE_CALL(cudaMallocHost(&host_buffer, size*sizeof(FLOAT)));
  CUDA_SAFE_CALL(cudaMalloc(&f, size*sizeof(FLOAT)));
  CUDA_SAFE_CALL(cudaMalloc(&fn, size*sizeof(FLOAT)));

  FLOAT  time      = 0.0;
  int    iter_count = 0;

  if (rank == 0) {
    startTimer();
    // copy data for rank0 self.
    memcpy(host_buffer, problem_f_, (count + rem_count + z_bound*2)*sizeof(FLOAT));
    for (int i = 1; i < nprocs; i++) {
      MPI_Send(problem_f_ + rem_count + i*count, count + z_bound*2,
               MPI_FLOAT, i, 0, MPI_COMM_WORLD);
    }
  }
  else {
    MPI_Recv(host_buffer, count + z_bound*2, MPI_FLOAT,
             0, 0, MPI_COMM_WORLD, &status);
  }

  CUDA_SAFE_CALL(cudaMemcpy(f, host_buffer, size*sizeof(FLOAT),
                            cudaMemcpyHostToDevice));

  dim3 grid(nx/DIM_X, ny/DIM_Y, 1);
  dim3 threads(DIM_X, DIM_Y, 1);
  FLOAT ce = k*dt/(dx*dx), cw = k*dt/(dx*dx),
    cn = k*dt/(dy*dy), cs = k*dt/(dy*dy),
    ct = k*dt/(dz*dz), cb = k*dt/(dz*dz),
    cc = 1.0 - (ce + cw + cn + cs + ct + cb);

  // run kernel
  do {
    MPI_Request send_req[2], recv_req[2];
    exchangeBoundaries(host_buffer, size, z_bound);
    if (rank > 0) {
      CUDA_SAFE_CALL(cudaMemcpy(f, host_buffer, z_bound*sizeof(FLOAT),
                                cudaMemcpyHostToDevice));
    }
    if (rank < nprocs - 1) {
      CUDA_SAFE_CALL(cudaMemcpy(f           + size - z_bound,
                                host_buffer + size - z_bound,
                                z_bound*sizeof(FLOAT),
                                cudaMemcpyHostToDevice));
    }

    if (iter_count && (iter_count % 100 == 0)) {
      LOG_INFO("time(%4d)=%7.5f", iter_count, time + dt);
    }

    mpi_gpu_diffusion3d<<<grid,threads>>>(f,fn,nx,ny,nz_self,
                                          ce,cw,cn,cs,ct,cb,cc);
    if (rank == 0) {
      CUDA_SAFE_CALL(cudaMemcpy(fn, fn + z_bound, z_bound*sizeof(FLOAT),
                                cudaMemcpyDeviceToDevice));
    }
    else if (rank == nprocs - 1) {
      CUDA_SAFE_CALL(cudaMemcpy(fn + size - z_bound,
                                fn + size - z_bound*2,
                                z_bound*sizeof(FLOAT),
                                cudaMemcpyDeviceToDevice));
    }

    // copy the boundaries to the host
    if (rank > 0) {
      CUDA_SAFE_CALL(cudaMemcpy(host_buffer + z_bound,
                                fn + z_bound,
                                z_bound*sizeof(FLOAT),
                                cudaMemcpyDeviceToHost));
    }
    if (rank < nprocs - 1) {
      CUDA_SAFE_CALL(cudaMemcpy(host_buffer + size - z_bound*2,
                                fn + size - z_bound*2,
                                z_bound*sizeof(FLOAT),
                                cudaMemcpyDeviceToHost));
    }

    std::swap(f, fn);
    time += dt;
    iter_count++;
  } while (time + 0.5*dt < 0.1);
  CUDA_SAFE_CALL(cudaMemcpy(host_buffer, f, size*sizeof(FLOAT),
                            cudaMemcpyDeviceToHost));

  // send data back to rank0
  if (rank == 0) {
    LOG_INFO("calculation has been done.");
    FLOAT *answer = new FLOAT[nx*ny*nz];
    memcpy(answer, host_buffer + z_bound, (count+rem_count)*sizeof(FLOAT));
    for (int i = 1; i < nprocs; i++) {
      MPI_Recv(answer + rem_count + i*count, count, MPI_FLOAT,
               i, 0, MPI_COMM_WORLD, &status);
    }
    LOG_INFO("recv checksum: %f", getSum(answer, nx*ny*nz));
    stopTimer();
    result_time_     = time;
    result_accuracy_ = problem_.accuracy(answer, time);
    result_flops_    = (nx*ny*nz)*13.0*iter_count/elapsed_time_*1.0e-06;
    delete[] answer;
  }
  else {
    MPI_Send(host_buffer + z_bound, count, MPI_FLOAT,
             0, 0, MPI_COMM_WORLD);
  }
  CUDA_SAFE_CALL(cudaFree(f));
  CUDA_SAFE_CALL(cudaFree(fn));
}
