#pragma once
#include "benchmark.h"

class MPI_Benchmark : public Benchmark {
protected:
  FLOAT   *fore_buffer_;
  FLOAT   *back_buffer_;
  FLOAT   *problem_f_;
  Size3d   problem_n_;
  int      mpi_rank_;
  int      mpi_num_process_;
  string   mpi_processor_name_;
  int      nz_self_;
  
  inline FLOAT getSum(FLOAT *f, int count) {
    FLOAT s = 0.0;
    while (count-- > 0) {
      s += *f++;
    }
    return s;
  }
  inline void exchangeBoundaries(FLOAT *f, int size, int z_bound) {
    MPI_Status status;
    MPI_Request send_req[2], recv_req[2];
    const int rank = mpi_rank_;
    const int nprocs = mpi_num_process_;
    if (rank > 0) {
      // send upper bound
      MPI_Isend(f + z_bound, z_bound, MPI_FLOAT,
                rank - 1, 0, MPI_COMM_WORLD, &send_req[0]);
      // recieve upper bound
      MPI_Irecv(f          , z_bound, MPI_FLOAT,
                rank - 1, 1, MPI_COMM_WORLD, &recv_req[1]);
    }
    if (rank < nprocs - 1) {
      // send bottom bound
      MPI_Isend(f + size - z_bound*2, z_bound, MPI_FLOAT,
                rank + 1, 1, MPI_COMM_WORLD, &send_req[1]);
      // recieve bottom bound
      MPI_Irecv(f + size - z_bound, z_bound, MPI_FLOAT,
                rank + 1, 0, MPI_COMM_WORLD, &recv_req[0]);
    }

    if (rank > 0) {
      MPI_Wait(&recv_req[1], &status);
    }
    if (rank < nprocs - 1) {
      MPI_Wait(&recv_req[0], &status);
    }
  }

  inline void diffusion3d(
    FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
    FLOAT dx, FLOAT dy, FLOAT dz, FLOAT dt, FLOAT kappa) {
    FLOAT ce, cw, cn, cs, ct, cb, cc;
    ce = cw = kappa*dt/(dx*dx);
    cn = cs = kappa*dt/(dy*dy);
    ct = cb = kappa*dt/(dz*dz);
    cc = 1.0 - (ce + cw + cn + cs + ct + cb);

    int j, jx, jy, jz;
    for (jz = 0; jz < nz; jz++) {
      for (jy = 0; jy < ny; jy++) {
        for (jx = 0; jx < nx; jx++) {
          FLOAT e, w, n, s, t, b, c;
          j = (jz+1)*nx*ny + jy*nx + jx;
          c = f[j];
          w = (jx == 0)    ? c : f[j-1];
          e = (jx == nx-1) ? c : f[j+1];
          n = (jy == 0)    ? c : f[j-nx];
          s = (jy == ny-1) ? c : f[j+nx];
          b = f[j-nx*ny];
          t = f[j+nx*ny];
          fn[j] = cc*c
            + cw*w + ce*e
            + cs*s + cn*n
            + cb*b + ct*t;
        }
      }
    }
  }

public:
  MPI_Benchmark() : fore_buffer_(NULL), back_buffer_(NULL), problem_f_(NULL) {
  }

  virtual ~MPI_Benchmark() {
    delete[] fore_buffer_;
    delete[] back_buffer_;
    delete[] problem_f_;
  }

  virtual void initialize(const int nx, const int ny, const int nz);
  virtual void run();
};

class MPI_GPUBenchmark : public MPI_Benchmark {
private:
public:
  MPI_GPUBenchmark() {}
  virtual void run();
};
