#include "types.h"
#include "problem.h"
#include "cuda_runtime.h"
#include "cutil.h"
#include "log.h"

__global__
void omp_gpu_diffusion3d(
        FLOAT *f, FLOAT *fn, int nx, int ny, int nz,
        FLOAT ce, FLOAT cw, FLOAT cn, FLOAT cs, FLOAT ct, FLOAT cb, FLOAT cc) {
    FLOAT e, w, n, s, t, b, c;
    int jx, jy, jz, j, ns;
    jx = blockDim.x*blockIdx.x + threadIdx.x;
    jy = blockDim.y*blockIdx.y + threadIdx.y;
    ns = nx*ny;

    j = jy*nx + jx + ns; // f and fn has a boundary buffer of size ns
                         // at the beginning 
    for (jz = 0; jz < nz; jz++) {
        int iw, ie, is, in, ib, it;
        c = f[j];
        iw = (jx == 0     ) ? j : j-1;
        ie = (jx == nx - 1) ? j : j+1;
        is = (jy == 0     ) ? j : j-nx;
        in = (jy == ny - 1) ? j : j+nx;
        ib = (jz == 0)    ? j : j-ns;
        it = (jz == nz-1) ? j : j+ns;

        w = f[iw];
        e = f[ie];
        s = f[is];
        n = f[in];
        b = f[ib];
        t = f[it];
        fn[j] = cc*c + cw*w + ce*e + cn*n + cs*s + cb*b + ct*t;
        j += ns;
    }
}

#define DIM_X 64
#define DIM_Y 4

void omp_gpu_run_diffusion3d(int tid, int nthreads,
                             Problem &prob,
                             FLOAT &final_time, 
                             int &final_iter_count)
{
    int    nx        = prob.n.x;
    int    ny        = prob.n.y;
    int    nz        = prob.n.z;
    int    ns        = nx * ny;
    FLOAT  dx        = prob.d.x;
    FLOAT  dy        = prob.d.y;
    FLOAT  dz        = prob.d.z;
    FLOAT  dt        = prob.dt;
    FLOAT  k         = prob.kappa;
    int    dim_z     = nz / nthreads;
    int    rem_z     = nz % nthreads;
    int    count     = ns * dim_z;
    int    z_bound   = ns;
    int    nz_self   = tid == 0 ? dim_z + rem_z : dim_z;
    int    size = ns * nz_self;
    dim3 grid(nx/DIM_X, ny/DIM_Y, 1);
    dim3 threads(DIM_X, DIM_Y, 1);
    FLOAT ce = k*dt/(dx*dx), cw = k*dt/(dx*dx),
            cn = k*dt/(dy*dy), cs = k*dt/(dy*dy),
            ct = k*dt/(dz*dz), cb = k*dt/(dz*dz),
            cc = 1.0 - (ce + cw + cn + cs + ct + cb);
    FLOAT *host_buffer = prob.f() + tid * count;
    FLOAT *f;
    FLOAT *fn;
    FLOAT  time      = 0.0;
    int    iter_count = 0;

    LOG_INFO("tid: %d", tid);
    CUDA_SAFE_CALL(cudaSetDevice(tid));
    CUDA_SAFE_CALL(cudaMalloc((void**)&f, (size + z_bound * 2) * sizeof(FLOAT)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&fn, (size + z_bound * 2) * sizeof(FLOAT)));

    CUDA_SAFE_CALL(cudaMemcpy(f + z_bound, host_buffer, size*sizeof(FLOAT),
                              cudaMemcpyHostToDevice));

    do {
        // Copy bottom boundary to device
        if (tid > 0) {
            CUDA_SAFE_CALL(cudaMemcpy(f, host_buffer - z_bound,
                                      z_bound*sizeof(FLOAT),
                                      cudaMemcpyHostToDevice));
        }
        // Copy top boundary to device
        if (tid < nthreads - 1) {
            CUDA_SAFE_CALL(cudaMemcpy(f + z_bound + size, host_buffer + size,
                                      z_bound * sizeof(FLOAT),
                                      cudaMemcpyHostToDevice));
        }

        if (iter_count && (iter_count % 100 == 0)) {
            LOG_INFO("time(%4d)=%7.5f", iter_count, time + dt);
        }

        omp_gpu_diffusion3d<<<grid,threads>>>(f,fn,nx,ny,nz_self,
                                              ce,cw,cn,cs,ct,cb,cc);

        // Copy bottom boundary to host        
        if (tid > 0) {
            CUDA_SAFE_CALL(cudaMemcpy(host_buffer, fn + z_bound,
                                      z_bound*sizeof(FLOAT),
                                      cudaMemcpyDeviceToHost));
        }
        // Copy top boundary to host                
        if (tid < nthreads - 1) {
            CUDA_SAFE_CALL(cudaMemcpy(host_buffer + size - z_bound,
                                      fn + size,
                                      z_bound*sizeof(FLOAT),
                                      cudaMemcpyDeviceToHost));
        }

        std::swap(f, fn);
        time += dt;
        iter_count++;
    } while (time + 0.5*dt < 0.1);
    
    if (tid == 0) {
        final_time = time;
        final_iter_count = iter_count;
    }
    // copy back the sub grid to host
    CUDA_SAFE_CALL(cudaMemcpy(host_buffer, f + z_bound, size * sizeof(FLOAT),
                              cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaThreadExit());
    return;
}


        
