/*
 * Original code written by Toshio Endo (endo@is.titech.ac.jp)
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define MAX_ITER (10000)
#define NX (16)
#define NY (16)

long time_diff_us(struct timeval st, struct timeval et)
{
    return (et.tv_sec-st.tv_sec)*1000000+(et.tv_usec-st.tv_usec);
}

int initialize(float *region[2], int nx, int ny)
{
    int rid;
    int i, j;
    /* initialize two regions */
    for (rid = 0; rid < 2; rid++) {
        for (i = 0; i < nx; i++) {
            for (j = 0; j < ny; j++) {
                region[rid][j*nx+i] = 0.0;
            }
        }
        /* right edge */
        for (j = 0; j < ny; j++) {
            region[rid][j*nx+(nx-1)] = (float)j/(float)ny;
        }
        /* lower edge */
        for (i = 0; i < nx; i++) {
            region[rid][(ny-1)*nx+i] = (float)i/(float)nx;
        }
    }
    return 0;
}

__global__ void jacobi(float *region1, float *region2, 
			int nx, int ny, float omega)
{
    int i, j;
    int iter;
    float *cur_region = region1;
    float *new_region = region2;
    float *tmp_region;

    for (iter = 0; iter < MAX_ITER; iter++) {
        for (j = 1; j < ny-1; j++) {
            for (i = 1; i < nx-1; i++) {
                float curv = cur_region[j*nx+i];
                float nextv;
                /* compute new value from neighbor points */
                nextv = (cur_region[(j-1)*nx+i] +
                         cur_region[(j+1)*nx+i] +
                         cur_region[j*nx+(i-1)] +
                         cur_region[j*nx+(i+1)]) * 0.25f;
                /* overcorrection */
                new_region[j*nx+i] = curv + (nextv - curv) * omega;
            }
        }
        /* switch buffer */
        tmp_region = cur_region;
        cur_region = new_region;
        new_region = tmp_region;
    }
    return;
}


int main(int argc, char *argv[])
{
    float omega = 0.5;
    float *region[2]; /* for double buffer */
    float *region_d[2]; /* for double buffer */    
    struct timeval st;
    struct timeval et;
    long us;

    /* allocate 2D region */
    /* double buffer */
    region[0] = (float*)malloc(sizeof(float)*NX*NY);
    region[1] = (float*)malloc(sizeof(float)*NX*NY);
    cudaMalloc((void **)&region_d[0], sizeof(float)*NX*NY);
    cudaMalloc((void **)&region_d[1], sizeof(float)*NX*NY);

    initialize(region, NX, NY);

    gettimeofday(&st, NULL); /* get start time */

    cudaMemcpy(region_d[0], region[0], sizeof(float)*NX*NY,
               cudaMemcpyHostToDevice);
    cudaMemcpy(region_d[1], region[1], sizeof(float)*NX*NY,
               cudaMemcpyHostToDevice);

    jacobi<<<1, 1>>>(region_d[0], region_d[1], NX, NY, omega);

    cudaMemcpy(region[0], region_d[0], sizeof(float)*NX*NY,
               cudaMemcpyDeviceToHost);
    cudaMemcpy(region[1], region_d[1], sizeof(float)*NX*NY,
               cudaMemcpyDeviceToHost);
    
    cudaThreadSynchronize();    
    gettimeofday(&et, NULL); /* get end time */

    us = time_diff_us(st, et);
    printf("Elapsed time: %ld us (%dx%d points, single GPU thread)\n",
           us, NX, NY);

    free(region[0]);
    free(region[1]);
    cudaFree(region_d[0]);
    cudaFree(region_d[1]);
    return 0;
}
